/**
 * OW2 Util
 * Copyright (C) 2008 Bull S.A.S.
 * Contact: easybeans@ow2.org
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
 *
 * --------------------------------------------------------------------------
 * $Id: CarDeployableMetadataFactory.java 4389 2008-12-15 13:48:57Z alitokmen $
 * --------------------------------------------------------------------------
 */

package org.ow2.util.ee.metadata.car.impl;

import java.net.URL;
import java.util.Arrays;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.ow2.util.archive.api.ArchiveException;
import org.ow2.util.archive.api.IArchive;
import org.ow2.util.ee.deploy.api.deployable.CARDeployable;
import org.ow2.util.ee.deploy.api.deployable.metadata.DeployableMetadataException;
import org.ow2.util.ee.metadata.car.api.ICarDeployableMetadata;
import org.ow2.util.ee.metadata.car.api.ICarDeployableMetadataFactory;
import org.ow2.util.ee.metadata.car.impl.configurator.CarDeployableMetadataConfigurator;
import org.ow2.util.ee.metadata.car.impl.specific.ISpecificCarDeployableFactory;
import org.ow2.util.ee.metadata.car.impl.xml.merge.CarMetadataMerge;
import org.ow2.util.ee.metadata.car.impl.xml.parsing.CARDeploymentDescLoader;
import org.ow2.util.ee.metadata.car.impl.xml.struct.CAR;
import org.ow2.util.ee.metadata.common.impl.AbsDeployableMetadataFactory;
import org.ow2.util.ee.metadata.common.impl.xml.parsing.ParsingException;
import org.ow2.util.log.Log;
import org.ow2.util.log.LogFactory;
import org.ow2.util.scan.api.IClassesLocator;
import org.ow2.util.scan.api.IScanner;
import org.ow2.util.scan.api.classlocator.ArchiveClassesLocator;
import org.ow2.util.scan.api.classlocator.ClassLoaderClassesLocator;

/**
 * Factory to contruct car deployable metadata.
 * @author Gael Lalire
 */
public class CarDeployableMetadataFactory extends AbsDeployableMetadataFactory implements ICarDeployableMetadataFactory {

    /**
     * Logger.
     */
    private static final Log LOGGER = LogFactory.getLog(CarDeployableMetadataFactory.class);

    /**
     * Application-client DD entry name.
     */
    private static final String APPLICATION_CLIENT = "META-INF/application-client.xml";

    /**
     * Constructor.
     */
    public CarDeployableMetadataFactory() {
    }

    /**
     * Constructor with an archive scanner factory.
     * @param scanner a class scanner
     */
    public CarDeployableMetadataFactory(final IScanner scanner) {
        super(scanner);
    }

    /**
     * Create car deployable metadata.
     * @param deployable the car deployable
     * @return the metadata
     * @throws DeployableMetadataException if error occurs
     */
    public ICarDeployableMetadata createDeployableMetadata(final CARDeployable deployable) throws DeployableMetadataException {
        return createDeployableMetadata(deployable, Arrays.asList(new ClassLoaderClassesLocator(ClassLoader
                .getSystemClassLoader())), null);
    }

    /**
     * Create car deployable metadata.
     * @param deployable the car deployable
     * @param classLoader a classloader to use if a class is not in archive
     * @return the metadata
     * @throws DeployableMetadataException if error occurs
     */
    public ICarDeployableMetadata createDeployableMetadata(final CARDeployable deployable, final ClassLoader classLoader)
            throws DeployableMetadataException {
        List<? extends IClassesLocator> classesLocatorList = null;
        if (classLoader != null) {
            classesLocatorList = Arrays.asList(new ClassLoaderClassesLocator(classLoader));
        } else {
            classesLocatorList = Arrays.asList(new ClassLoaderClassesLocator(ClassLoader.getSystemClassLoader()));
        }
        return createDeployableMetadata(deployable, classesLocatorList, null);
    }

    /**
     * Extract a class list from xml.
     * @param car representation of application-client.xml
     * @param classNameSet a set to fill with classes name referenced in application-client.xml
     */
    private void completeClassList(final CAR car, final Set<String> classNameSet) {
        if (car != null) {
            addClassToSetIfNotNull(car.getCallbackHandler(), classNameSet);
        }

        //TODO: complete with all elements like all ejb-ref, resource-ref, etc
    }

    /**
     * Create car deployable metadata.
     * @param deployable the car deployable
     * @param carExternLibraries extern libraries
     * @param specificDeployableFactoryList a specific deployable factory list
     * @return the metadata
     * @throws DeployableMetadataException if error occurs
     */
    public ICarDeployableMetadata createDeployableMetadata(final CARDeployable deployable,
            final List<? extends IClassesLocator> carExternLibraries,
            final List<? extends ISpecificCarDeployableFactory> specificDeployableFactoryList)
            throws DeployableMetadataException {

        // for time debugging
        long tAnalyzeStart = System.currentTimeMillis();

        IArchive archive = deployable.getArchive();

        CAR car = null;
        URL applicationClientURL;
        try {
            applicationClientURL = archive.getResource(APPLICATION_CLIENT);
        } catch (ArchiveException e) {
            throw new DeployableMetadataException("Unable to get DD URL", e);
        }

        if (applicationClientURL != null) {
            try {
                car = CARDeploymentDescLoader.loadDeploymentDescriptor(applicationClientURL);
            } catch (ParsingException e) {
                throw new DeployableMetadataException("Unable to parse DD URL", e);
            }
        }

        CarDeployableMetadataConfigurator carArchiveMetadataConfigurator = new CarDeployableMetadataConfigurator(deployable);


        Set<String> classNameSet = new HashSet<String>();

        if (specificDeployableFactoryList != null) {
            for (ISpecificCarDeployableFactory specificEjbJarDeployableFactory : specificDeployableFactoryList) {
                carArchiveMetadataConfigurator.addSpecificMetadataConfigurator(specificEjbJarDeployableFactory
                        .beforeScan(deployable, classNameSet));
            }
        }

        // if metadata is complete then filter class to scan
        if (car != null && car.isMetadataComplete()) {
            // disable annotation parsing
            carArchiveMetadataConfigurator.setAnnotationParsingDesactived(true);

        } else {
            // enable annotation parsing because metadata descriptor is not
            // complete
            carArchiveMetadataConfigurator.setAnnotationParsingDesactived(false);
        }

        // FIXME: this should be done only if metadata is completed
        Iterator<String> entryIterator;
        try {
            entryIterator = archive.getEntries();
        } catch (ArchiveException e) {
            throw new DeployableMetadataException(e);
        }
        while (entryIterator.hasNext()) {
            String entry = entryIterator.next();
            if (entry.endsWith(CLASS_SUFFIX)) {
                classNameSet.add(entry.substring(0, entry.length() - CLASS_SUFFIX.length()));
            }
        }
        // END FIXME


        // get class name to scan from deployment descriptor
        completeClassList(car, classNameSet);

        // do annotation scan
        scan(classNameSet, new ArchiveClassesLocator(archive), carExternLibraries, carArchiveMetadataConfigurator);

        // get result
        CarDeployableMetadata carDeployableMetadata = carArchiveMetadataConfigurator.getCarDeployableMetadata();

        // Default is "not completed"
        if (car != null) {
            carDeployableMetadata.setMetadataComplete(car.isMetadataComplete());
        }

        // do merge with DD
        if (car != null) {
            new CarMetadataMerge(carDeployableMetadata).merge(car);
        }

        // time if debugging
        if (LOGGER.isDebugEnabled()) {
            long tAnalyzeStartEnd = System.currentTimeMillis();
            LOGGER.debug("Analyze of file {0} took {1} ms.", archive.getName(), (tAnalyzeStartEnd - tAnalyzeStart));
        }

        return carDeployableMetadata;
    }

    /**
     * @return the deployable metadata class
     */
    public Class<ICarDeployableMetadata> getDeployableMetadataClass() {
        return ICarDeployableMetadata.class;
    }

}
